import 'dart:convert';
import 'dart:typed_data';
import 'package:http/http.dart' as http;
import 'package:http/http.dart';
import 'package:modular_system_core/mscore.dart';

export './http_requester.dart';

class HttpRequester {
  Future<HttpResponse> sendAsync(HttpRequest request) {
    switch (request.method) {
      case HttpMehotd.get:
        return getAsync(request);
      case HttpMehotd.post:
        return postAsync(request);
      case HttpMehotd.put:
        return putAsync(request);
      case HttpMehotd.patch:
        return patchAsync(request);
      case HttpMehotd.delete:
        return deleteAsync(request);
      default:
        throw AppException("Invalid HttpMehtod at HttpRequester.", ExceptionCode.internal, StackTrace.current);
    }
  }

  Future<HttpResponse> getAsync(HttpRequest request) async {
    var resp = await http.get(request.uri);
    return adaptResponse(request, resp);
  }

  Future<HttpResponse> postAsync(HttpRequest request) async {
    var resp = await http.post(
      request.uri,
      headers: request.headers,
      body: request.body.value(),
    );
    return adaptResponse(request, resp);
  }

  Future<HttpResponse> putAsync(HttpRequest request) async {
    var resp = await http.put(
      request.uri,
      headers: request.headers,
      body: request.body.value(),
    );
    return adaptResponse(request, resp);
  }

  Future<HttpResponse> patchAsync(HttpRequest request) async {
    var resp = await http.patch(
      request.uri,
      headers: request.headers,
      body: request.body.value(),
    );
    return adaptResponse(request, resp);
  }

  Future<HttpResponse> deleteAsync(HttpRequest request) async {
    var resp = await http.delete(
      request.uri,
      headers: request.headers,
      body: request.body.value(),
    );
    return adaptResponse(request, resp);
  }

  Future<HttpResponse> optionsAsync(HttpRequest request) async {
    throw UnimplementedError();
  }

  HttpResponse adaptResponse(HttpRequest request, Response originalResponse) {
    return HttpResponse(
      statusCode: originalResponse.statusCode,
      uri: request.uri,
      method: request.method,
      body: ByteArrayBody(
        data: originalResponse.bodyBytes,
        contentType: originalResponse.headers["content-type"] ?? "",
      ),
      request: request,
      originalResponse: originalResponse,
    );
  }
}

enum HttpMehotd {
  get,
  post,
  put,
  patch,
  delete,
  options,
}

// components
abstract class HttpBody {
  String contentType;
  Uint8List? data;

  HttpBody({
    this.contentType = "text/plain",
    this.data,
  });

  Object? value();
}

class EmptyBody extends HttpBody {
  @override
  Object? value() {
    return null;
  }
}

class ByteArrayBody extends HttpBody {
  ByteArrayBody({
    required Uint8List data,
    required String contentType,
  }) : super(data: data, contentType: contentType);

  @override
  Object? value() {
    return data;
  }
}

class JsonBody extends HttpBody {
  String json;

  JsonBody(this.json) : super(contentType: "application/json");

  @override
  Object? value() {
    return json;
  }
}

class TypedJsonBody<T extends JsonSerializer<T>> extends HttpBody {
  T type;

  TypedJsonBody(this.type) : super(contentType: "application/json");

  @override
  Object? value() {
    return type.toJson();
  }
}

abstract class RestfulEndpoint<InputT, OutputT> {
  HttpRequester requester = HttpRequester();

  Future<OutputT> runAsync(InputT input) async {
    var req = getRequest(input);
    var resp = await requester.sendAsync(req);

    if (resp.isSuccess()) {
      return await deserializeResponseAsync(resp);
    }
    throw handleFailure(resp);
  }

  HttpRequest getRequest(InputT input);
  Future<OutputT> deserializeResponseAsync(HttpResponse response);
  Exception handleFailure(HttpResponse response);
}

// request
class HttpRequest {
  Uri uri;
  HttpMehotd method;
  HttpBody body;
  Map<String, String> headers;

  HttpRequest({
    required this.uri,
    required this.method,
    required this.body,
    this.headers = const {},
  });
}

// response
class HttpResponse {
  int statusCode;
  Uri uri;
  HttpMehotd method;
  HttpBody body;
  Map<String, String> headers;
  HttpRequest request;
  Response originalResponse;

  HttpResponse({
    required this.statusCode,
    required this.uri,
    required this.method,
    required this.body,
    required this.request,
    required this.originalResponse,
    this.headers = const {},
  });

  bool isSuccess() {
    return statusCode >= 200 && statusCode <= 299;
  }

  String readAsUtf8() {
    if (body.data != null) {
      return utf8.decode(body.data!, allowMalformed: true);
    }
    return "";
  }

  Map<String, dynamic>? readAsJsonMap() {
    return jsonDecode(readAsUtf8());
  }
}
