import 'package:modular_system_core/src/extension.dart';

import '../models.dart';
import 'delegates.dart';
import 'http.dart';

class ApiClient {
  late Uri uri;
  late Function(HttpOutput output)? validationCallback;
  late ConcreteHttp client = const ConcreteHttp();

  ApiClient({
    required this.uri,
    this.validationCallback,
  });

  Future<HttpOutput> sendAsync(HttpInput input, {Function(HttpOutput output)? validationCallback}) async {
    var output = await client.sendAsync(input);

    if (validationCallback != null) {
      validationCallback(output);
    } else {
      okCallback(output);
    }
    return output;
  }

  void okCallback(HttpOutput output) {
    if (!output.isOk) {
      throw AppException.fromJson(output.message ?? "The request could not be fulfilled.", StackTrace.current);
    }
  }

  void jsonCallback(HttpOutput output) {
    if (output.asJsonDecoded() == null) {
      throw AppException.fromJson(output.message ?? "Could not deserialize JSON body.", StackTrace.current);
    }
  }

  void okJsonCallback(HttpOutput output) {
    okCallback(output);
    jsonCallback(output);
  }
}

class CrudClient<T extends JsonSerializer<T>> extends ApiClient {
  CrudClient({
    required Uri uri,
    required this.controllerName,
    required this.constructor,
    this.token,
  }) : super(uri: uri) {
    setSessionToken();
  }

  String controllerName;
  String? token;
  JsonObjectCreator<T> constructor;

  void setSessionToken() {
    if (token != null) {
      uri = uri.addQueryParam("app-session", token!);
    }
  }

  // CREATE
  /// Creates an instance of T and returns its ID.
  Future<StringDTO> createAsync(T data) async {
    var req = HttpInput(
      uri: uri.replace(
        path: "api/$controllerName",
      ),
      method: HttpMethod.post,
      body: HttpBody.json(data.toJson()),
    );
    var resp = await sendAsync(req, validationCallback: okJsonCallback);
    var json = resp.asJsonMap();
    return StringDTO.fromJson(json!);
  }

  // READ
  Future<T> getByIdAsync(String id) async {
    var req = HttpInput(
      uri: uri.replace(
        path: "api/$controllerName",
        queryParameters: {"id": id},
      ),
      method: HttpMethod.get,
    );
    var resp = await sendAsync(req, validationCallback: okJsonCallback);
    var json = resp.asJsonMap();
    return constructor(json!);
  }

  Future<QueryList<T>> searchAsync(WebSearchRequest search) async {
    var req = HttpInput(
      uri: uri.replace(
        path: "api/$controllerName/query",
      ),
      method: HttpMethod.post,
      body: HttpBody.json(search.toJson()),
    );
    var resp = await sendAsync(req, validationCallback: okJsonCallback);
    var json = resp.asJsonMap();
    return QueryList.fromJsonMap(json!, constructor);
  }

  // UPDATE
  Future<BooleanDTO> updateAsync(WebUpdateRequest update) async {
    var req = HttpInput(
      uri: uri.replace(
        path: "api/$controllerName",
      ),
      method: HttpMethod.put,
      body: HttpBody.json(update.toJson()),
    );
    var resp = await sendAsync(req, validationCallback: okJsonCallback);
    var json = resp.asJsonMap();
    return BooleanDTO.fromJson(json!);
  }

  Future<BooleanDTO> overrideAsync(T data) async {
    var req = HttpInput(
      uri: uri.replace(
        path: "api/$controllerName/override",
      ),
      method: HttpMethod.put,
      body: HttpBody.json(data.toJson()),
    );
    var resp = await sendAsync(req, validationCallback: okJsonCallback);
    var json = resp.asJsonMap();
    return BooleanDTO.fromJson(json!);
  }

  // DELETE
  Future<BooleanDTO> deleteByIdAsync(String id) async {
    var req = HttpInput(
      uri: uri.replace(
        path: "api/$controllerName",
        queryParameters: {"id": id},
      ),
      method: HttpMethod.delete,
    );
    var resp = await sendAsync(req, validationCallback: okJsonCallback);
    var json = resp.asJsonMap();
    return BooleanDTO.fromJson(json!);
  }
}
